#!/bin/bash

#Run arguments if provided 
if [ -n "$@" ]; then exec $@; fi

export LOG_REDIR=${LOG_REDIR:-/dev/stdout}
SHUTDOWN=0
DXLPID=0

function set_ready() {
	if [ -n "$READY_FILE" ]; then echo "$1" > "$READY_FILE"; fi
}

function do_setup () {
	FIRSTRUN="/etc/dh2i/firstrun"
	READY_FILE="/opt/dh2i/sbin/ready"
	HN=`uname -n`

	# Set defaults
	DX_VHOST_NAME=${DX_VHOST_NAME:-VHOST1}
	DX_VHOST_NAME=${DX_VHOST_NAME^^}
	if [ -z "$DX_AG_NAME" -a -z "$DX_PGR_NAME" ]; then DX_AG_NAME=AG1; fi
	DX_AG_NAME=${DX_AG_NAME^^}
	DX_PGR_NAME=${DX_PGR_NAME^^}
	PGDATA=${PGDATA:-/var/lib/postgresql/data}
	PGPORT=${PGPORT:-5432}
	PGUSER=${PGUSER:-postgres}
	DX_NEW_CLUSTER=${DX_NEW_CLUSTER:-true}
	DX_NEW_CLUSTER=${DX_NEW_CLUSTER,,}
	# For backward-compatibility, do NAT joins if neither DX_USE_NAT and DX_JOIN_TARGET are set, but DX_OTPK is set
	# If DX_JOIN_TARGET is set, but DX_USE_NAT is not set, presume no NAT
	if [ -z "$DX_USE_NAT" -a -z "$DX_JOIN_TARGET" -a -n "$DX_OTPK" ]; then DX_USE_NAT=true; fi
	DX_USE_NAT=${DX_USE_NAT:-false}
	DX_USE_NAT=${DX_USE_NAT,,}
	DX_OPERATOR=${DX_OPERATOR:-false}
	DX_USE_TUNNELS=${DX_USE_TUNNELS:-true}

	# Check for firstrun file
	if [ -f "$FIRSTRUN" ]; then 
		echo "Container has an existing configuration, skipping setup." >>$LOG_REDIR 2>&1
		set_ready 1
		return
	fi

	if [ -z "$DX_PASSKEY" ]; then 
		echo "No cluster passkey provided." >>$LOG_REDIR 2>&1
	else
		dxcli cluster-set-secret-ex "$DX_PASSKEY" >>$LOG_REDIR 2>&1
		if [ $? -ne 0 ]; then 
			echo "Failed to set cluster secret." >>$LOG_REDIR 2>&1
			set_ready 0
			return
		fi
	fi

	# If this is a pod/container managed by DxOperator, don't do setup
	if $DX_OPERATOR; then 
		set_ready 0
		touch "$FIRSTRUN"
		return
	fi

	if [ -z "$DX_ACCEPT_EULA" ]; then
		echo "You must accept the EULA to continue." >>$LOG_REDIR 2>&1
		set_ready 1
		return
	fi

	if [ ! -f /etc/dh2i/DxLicense.xml ]; then
		if [ -z "$DX_LICENSE" ]; then
			echo "No DxEnterprise license key provided, skipping setup." >>$LOG_REDIR 2>&1
			set_ready 0
			return
		else
			dxcli activate-server --accept-eula "$DX_LICENSE" >>$LOG_REDIR 2>&1
			if [ $? -ne 0 ]; then 
				echo "License activation failed.". >>$LOG_REDIR 2>&1
				set_ready 0
				return
			fi
		fi
	fi

	# Prepare the primary for cluster joins if this is a new cluster
	if [[ $HN == *"-0" ]] && [ "$DX_NEW_CLUSTER" != "false" ]; then 
		dxcli set-globalsetting "membername.lookup" "true" >>$LOG_REDIR 2>&1
		if [ -n "$DX_MATCHMAKER" -a "$DX_MATCHMAKER" != "match.dh2i.com" ]; then
			dxcli set-globalsetting "nat.agent" "${DX_MATCHMAKER}" >>$LOG_REDIR 2>&1
		fi
		if [ -z "$DX_OTPK" ]; then
			echo "No OTPK was provided. You must provide an OTPK so other members may join the cluster." >>$LOG_REDIR 2>&1
		else
			dxcli set-otpk "" "$DX_OTPK" >>$LOG_REDIR 2>&1
			dxcli set-globalsetting "membername.alerts" "false" >>$LOG_REDIR 2>&1
			if [ $? -ne 0 ]; then
				echo "Failed to set the cluster OTPK." >>$LOG_REDIR 2>&1
				set_ready 0
				return
			fi
		fi

		touch "$FIRSTRUN"

		# Create the Vhost
		dxcli cluster-add-vhost "$DX_VHOST_NAME" "*127.0.0.1" "$HN"  >>$LOG_REDIR 2>&1
		if [ $? -ne 0 ]; then
			echo "Failed to create Vhost." >>$LOG_REDIR 2>&1
			set_ready 0
			return
		fi

	# Otherwise, join the cluster
	else
		if [ "$DX_USE_NAT" = false ]; then
			# Direct join
			if [ -z "$DX_PASSKEY" -a -z "$DX_OTPK" ]; then
				echo "No cluster passkey or OTPK has been provided. Skipping join." >>$LOG_REDIR 2>&1
				set_ready 1
				touch "$FIRSTRUN"
				return;
			fi
			# If DX_JOIN_TARGET is not set, presume Kubernetes pod naming, and infer the -0 pod
			DX_JOIN_TARGET=${DX_JOIN_TARGET:-${HN%-*}-0}
			a=0
			while [ $a -lt 9 ]; do
				dxcli join-cluster-ex "$DX_JOIN_TARGET" "${DX_OTPK:-${DX_PASSKEY}}" >>$LOG_REDIR 2>&1
				if [ $? -eq 0 ]; then break; fi
				a=$((a+1))
				echo "Join attempts: $a" >>$LOG_REDIR 2>&1
				sleep 10
			done

			if [ $a -ge 9 ]; then 
				echo "Failed to join \'$DX_JOIN_TARGET\'." >>$LOG_REDIR 2>&1
				set_ready 0
				return
			fi
		else
			# NAT join
			if [ -z "$DX_OTPK" ]; then
				echo "NAT join is requested and no OTPK has been provided. Skipping join." >>$LOG_REDIR 2>&1
				set_ready 1
				touch "$FIRSTRUN"
				return;
			fi
			# If DX_JOIN_TARGET is not set, use DX_MATCHMAKER
			# If neither are set, use the backend default
			DX_JOIN_TARGET=${DX_JOIN_TARGET:-${DX_MATCHMAKER}}
			a=0
			while [ $a -lt 9 ]; do
				dxcli join-cluster-ex "$DX_JOIN_TARGET" "$DX_OTPK" "true" >>$LOG_REDIR 2>&1
				if [ $? -eq 0 ]; then break; fi

				a=$((a+1))
				echo "Join attempts: $a" >>$LOG_REDIR 2>&1
				sleep 10
			done

			if [ $a -ge 9 ]; then 
				echo "Failed to join cluster using the provided OTPK." >>$LOG_REDIR 2>&1
				set_ready 0
				return
			fi
		fi

		touch "$FIRSTRUN"
	fi

	if [ -n "$DX_AG_NAME" ]; then
	    # Check SA password
	    if [ -z "$MSSQL_SA_PASSWORD" ] && [ -z "$SA_PASSWORD" ]; then
		    echo "Must provide a password for SQL Server to create an availability group." >>$LOG_REDIR 2>&1
		    set_ready 1
		    return
	    else
		    if [ -z "$MSSQL_SA_PASSWORD" ]; then MSSQL_SA_PASSWORD=$SA_PASSWORD; fi
		    SQLPW=`dxcli encrypt-text "$MSSQL_SA_PASSWORD"`
		    if [ $? -ne 0 ]; then 
			    echo "Failed to encrypt the SA password." >>$LOG_REDIR 2>&1
			    set_ready 0
			    return
		    fi
	    fi

	    TIMEOUT=$((SECONDS+600))
		TUNNEL_OPT=""
		if [ "$DX_USE_TUNNELS" = true ]; then
			TUNNEL_OPT="|0"
		fi
	    while true; do
			# Try to create or join the availability group
			if [[ $HN == *"-0" ]] && [ "$DX_NEW_CLUSTER" != "false" ]; then 
				echo "Creating the availability group." >>$LOG_REDIR 2>&1
				dxcli add-ags "$DX_VHOST_NAME" "$DX_AG_NAME" "$HN|MSSQLSERVER|SA|$SQLPW|5022|SYNCHRONOUS_COMMIT$TUNNEL_OPT" "$DX_AG_OPTIONS" >>$LOG_REDIR 2>&1
			else
				echo "Adding $HN to the availability group." >>$LOG_REDIR 2>&1
				dxcli add-ags-node "$DX_VHOST_NAME" "$DX_AG_NAME" "$HN|MSSQLSERVER|SA|$SQLPW|5022|SYNCHRONOUS_COMMIT$TUNNEL_OPT" >>$LOG_REDIR 2>&1
			fi

			if [ $? -ne 0 ]; then
				if [ $SECONDS -gt $TIMEOUT ]; then
					echo "Failed to setup availability group." >>$LOG_REDIR 2>&1
					set_ready 0
					return
				fi

				sleep 15

			else
				echo "Successfully added availability group." >>$LOG_REDIR 2>&1
				break
			fi
	    done
	fi

	if [ -n "$DX_PGR_NAME" ]; then
	    if [ ! -f "$PGDATA/postmaster.opts" ] ; then
			echo "Unable to read PID file for PostgreSQL server" >>$LOG_REDIR 2>&1
			set_ready 0
			return
		fi
	    PG_CMDL=`cat "$PGDATA/postmaster.opts"`
	    set "$PG_CMDL"
	    PG_BINPATH=`dirname "$1"`
	    PGR_SELF_DESC="$HN|$PG_BINPATH|XXX|$PGDATA|"
	    TIMEOUT=$((SECONDS+600))
		ADDED=0
	    while true; do
		if [ ! -f "$PGDATA/postmaster.opts" ]; then false; else
			# Try to create or join the availability group
			if [[ $HN == *"-0" ]] && [ "$DX_NEW_CLUSTER" != "false" ]; then 
				echo "Creating the PosgreSQL Replication set." >>$LOG_REDIR 2>&1
				dxcli add-pgsql-replication "$DX_VHOST_NAME" "$DX_PGR_NAME" "$PGUSER" "$PGPASSWORD" "" "" "$PGPORT" "0" "$PGR_SELF_DESC" >>$LOG_REDIR 2>&1
			else
				true
				if [ $ADDED -eq 0 ]; then
					echo "Adding $HN to the PostgreSQL replication set." >>$LOG_REDIR 2>&1
					dxcli set-pgsql-replication-nodeparam "$DX_VHOST_NAME" "$DX_PGR_NAME" "$PGR_SELF_DESC" >>$LOG_REDIR 2>&1
					if [ $? -eq 0 ]; then
						dxcli add-node "$DX_VHOST_NAME" "$HN" >>$LOG_REDIR 2>&1
						ADDED=1
					fi
				fi
				# Allow the backup operation to be retried
				if [ $? -eq 0 ]; then
					echo "Seeding $HN from replication primary." >>$LOG_REDIR 2>&1
					dxcli backup-pgsql "$DX_VHOST_NAME" "$DX_PGR_NAME" "$HN" >>$LOG_REDIR 2>&1
				fi
			fi
		fi

		if [ $? -ne 0 ]; then
			if [ $SECONDS -gt $TIMEOUT ]; then
				echo "Failed to setup PostgreSQL replication set." >>$LOG_REDIR 2>&1
				set_ready 0
				return
			fi

			sleep 15
		else
			echo "Successfully added PostgreSQL replication set." >>$LOG_REDIR 2>&1
			break
		fi
	    done
	fi

	set_ready 1
	return
}

function container_shutdown() {
	# Find DxLMonitor and send it SIGTERM
	echo "Container shutting down..." >>$LOG_REDIR 2>&1
	SHUTDOWN=1
	if [ $DXLPID -ne 0 ]; then
		echo "Signalling DxLMonitor ($DXLPID) to shut down" >>$LOG_REDIR 2>&1
		kill -TERM $DXLPID
	fi
}
trap container_shutdown TERM INT QUIT HUP

set_ready 0

if ! [ -z $DX_HAS_MSSQLSERVER ] && [ $DX_HAS_MSSQLSERVER -eq 1 ]; then
	#Enable HADR and get SQL version
	export MSSQL_ENABLE_HADR=1
	export sqlServerVersion_MSSQLSERVER=$(./opt/mssql/bin/sqlservr -v | grep '^.*\([0-9]*\.\)\{3\}[0-9]*.*$' | head -1 | sed 's/^.* \(\([0-9]*\.\)\{3\}[0-9]*\).*$/\1/')

	i=0
	#Start SQL Server and wait for listener
	/opt/mssql/bin/sqlservr >>$LOG_REDIR 2>&1 &
	while ! grep -q "00000000:`printf "%04X" ${MSSQL_TCP_PORT:-1433}`" /proc/net/tcp; do
		if [ -z "`jobs -r`" ]; then exit 1; fi
		if [ "$i" -ge 300 ]; then
			cat /proc/net/tcp >>$LOG_REDIR 2>&1
			echo "SQL Server did not complete startup with open listening ports, aborting." >>$LOG_REDIR 2>&1
			exit 1
		fi
		sleep 1
		i=$(($i+1))
	done
fi

i=0
#Start DxEnterprise
/opt/dh2i/sbin/DxCMonitor run-service -f >>$LOG_REDIR 2>&1 &
/opt/dh2i/sbin/DxLEngine run-service -f >>$LOG_REDIR 2>&1 &
/opt/dh2i/sbin/DxHealthMonitor run-service -f >>$LOG_REDIR 2>&1 &

#Check services are listening on assigned ports
while true; do
	if grep -q "0100007F:`printf "%04X" ${DX_DXLENGINE_LISTENPORT:-7804}`" /proc/net/tcp && \
		grep -q "00000000:`printf "%04X" ${DX_DXCMONITOR_LISTENPORT:-7980}`" /proc/net/tcp && \
		grep -q "0100007F:`printf "%04X" ${DX_DXHEALTHMONITOR_LISTENPORT:-7802}`" /proc/net/tcp; then break; fi

    if [ "$i" -ge 60 ]; then
		cat /proc/net/tcp >>$LOG_REDIR 2>&1
		if ! grep -q "0100007F:`printf "%04X" ${DX_DXHEALTHMONITOR_LISTENPORT:-7802}`" /proc/net/tcp; then
			echo "DxHealthMonitor did not complete startup with open listening ports, aborting." >>$LOG_REDIR 2>&1
		fi
		if ! grep -q "0100007F:`printf "%04X" ${DX_DXLENGINE_LISTENPORT:-7804}`" /proc/net/tcp; then
			echo "DxLEngine did not complete startup with open listening ports, aborting." >>$LOG_REDIR 2>&1
		fi
		if ! grep -q "00000000:`printf "%04X" ${DX_DXCMONITOR_LISTENPORT:-7980}`" /proc/net/tcp; then
			echo "DxCMonitor did not complete startup with open listening ports, aborting." >>$LOG_REDIR 2>&1
		fi
		exit 1
	fi
	sleep 1
	i=$(($i+1))
done

# Start DxLMonitor
# This process is the main process; when it terminates, we terminate all others
/opt/dh2i/sbin/DxLMonitor run-service -f >>$LOG_REDIR 2>&1 &
DXLPID=$!

# Wait for DxL to Start
while ! grep -q "00000000:`printf "%04X" ${DX_DXLMONITOR_LISTENPORT:-7979}`" /proc/net/tcp; do
	if [ "$i" -ge 60 ]; then
		cat /proc/net/tcp >>$LOG_REDIR 2>&1
		echo "DxLMonitor did not complete startup with open listening ports, aborting." >>$LOG_REDIR 2>&1
		exit 1
	fi
	sleep 1
	i=$(($i+1))
done

# If we caught the shutdown signal already...
if [ $SHUTDOWN -ne 0 ]; then kill -TERM $DXLPID; fi

# Wait for DxLMonitor to exit
echo "DxLMonitor pid is $DXLPID" >>$LOG_REDIR 2>&1
do_setup
wait $DXLPID
EXITCODE=$?
# Wait can be interrupted by a signal, and isn't auto-restarted
# In this case, it will exit with code 143.  ??!!
if [ $EXITCODE -eq 143 ]; then
	trap - TERM INT QUIT HUP
	wait $DXLPID
	EXITCODE=$?
fi
echo "DxLMonitor has exited ($EXITCODE)" >>$LOG_REDIR 2>&1

# Send SIGTERM to all other jobs
XJOBS=`jobs -p`
if [ -n "$XJOBS" ]; then
	echo -n "Signalling termination to auxiliary jobs:" >>$LOG_REDIR 2>&1
	for pid in $XJOBS; do
		echo -n " [$pid]" >>$LOG_REDIR 2>&1
		kill -TERM $pid
	done
	echo
	echo "Waiting for jobs to exit..." >>$LOG_REDIR 2>&1
	wait
fi

echo "Container shutdown complete." >>$LOG_REDIR 2>&1

# Return the DxLMonitor exit code
exit $EXITCODE
